#include <windows.h>

#include "../Daodan.h"
#include "../Daodan_Config.h"
#include "Utility.h"
#include "Input.h"
#include "Win32.h"

#include "../Oni/Oni.h"

// How many lines to scroll per-WHEEL_DELTA, from SPI_GETWHEELSCROLLLINES
static int DDgPlat_WheelScrollLines = 3;

// High-DPI awareness functions
static BOOL (WINAPI *pAdjustWindowRectExForDpi)(LPRECT lpRect, DWORD dwStyle,
	BOOL bMenu, DWORD dwExStyle, UINT dpi) = NULL;
static UINT (WINAPI *pGetDpiForWindow)(HWND hwnd) = NULL;

// LIrPlatform_Terminate wrapper. Removes any cursor clipping we've done. Doing
// this is required for Windows 98 :-D
void ONICALL DD_LIrPlatform_Terminate(void)
{
	ClipCursor(NULL);
	LIrPlatform_Terminate();
}

// LIrPlatform_Mode_Set wrapper. Clips cursor to window bounds to
// prevent loosing focus (mostly on Linux).
void ONICALL DD_LIrPlatform_Mode_Set(unsigned int active_mode)
{
	DDmAssert(ONgPlatformData.Window);

	if (active_mode)
	{
		RECT rc;
		POINT pt;

		pt.x = 0;
		pt.y = 0;

		if (GetClientRect(ONgPlatformData.Window, &rc) &&
			ClientToScreen(ONgPlatformData.Window, &pt))
		{
			rc.left   += pt.x;
			rc.top    += pt.y;
			rc.right  += pt.x;
			rc.bottom += pt.y;

			ClipCursor(&rc);
		}
	}
	else
	{
		ClipCursor(NULL);
	}

	LIrPlatform_Mode_Set(active_mode);
}

BOOL WINAPI DD_GetCursorPos(LPPOINT lpPoint)
{
	DDmAssert(ONgPlatformData.Window);
	
	return GetCursorPos(lpPoint) && ScreenToClient(ONgPlatformData.Window, lpPoint);
}

BOOL WINAPI DD_SetCursorPos(int X, int Y)
{
	POINT pt;
	pt.x = X;
	pt.y = Y;
	
	DDmAssert(ONgPlatformData.Window);
	
	return ClientToScreen(ONgPlatformData.Window, &pt) && SetCursorPos(pt.x, pt.y);
}

static void HandleClose(void)
{
	// There's no way to reliably terminate a modal dialog.
	// The following condition is (almost) always true.
	if (WMgActive)
		exit(0);

	ONgTerminateGame = UUcTrue;
}

void DDrAdjustWindowRect_WithStyle(RECT *rc, DWORD style, DWORD exstyle,
                                   DWORD dpi)
{
	HWND window = ONgPlatformData.Window;

	if (DDgDpiAwareness == DPI_AWARENESS_CONTEXT_PER_MONITOR_AWARE_V2) {
		if (!dpi)
			dpi = pGetDpiForWindow(window);
		pAdjustWindowRectExForDpi(rc, style, FALSE, exstyle, dpi);
	} else {
		AdjustWindowRectEx(rc, style, FALSE, exstyle);
	}
}

static void DDrAdjustWindowRect(RECT *rc, DWORD dpi)
{
	HWND window = ONgPlatformData.Window;
	DWORD style = GetWindowLongPtrA(window, GWL_STYLE);
	DWORD exstyle = GetWindowLongPtrA(window, GWL_EXSTYLE);
	DDrAdjustWindowRect_WithStyle(rc, style, exstyle, dpi);
}

static void HandleDpiChanged(HWND window, UINT dpi, RECT *suggested_rc)
{
	// Get the current physical size of the window and add window borders for
	// the new DPI
	RECT rc;
	GetClientRect(window, &rc);
	DDrAdjustWindowRect(&rc, dpi);

	// Is the window currently being moved?
	GUITHREADINFO ti = { .cbSize = sizeof ti };
	GetGUIThreadInfo(GetCurrentThreadId(), &ti);
	bool in_move_size = ti.hwndMoveSize == window;

	RECT wnd_rc;
	RECT *ref_rc;
	if (in_move_size) {
		// If the window is currently being moved, center it on its old size.
		// This is to keep it under the mouse cursor.
		GetWindowRect(window, &wnd_rc);
		ref_rc = &wnd_rc;
	} else {
		// Otherwise, center it on the new suggested rectangle
		ref_rc = suggested_rc;
	}
	OffsetRect(&rc,
		(ref_rc->left + ref_rc->right) / 2 - (rc.left + rc.right) / 2,
		(ref_rc->top + ref_rc->bottom) / 2 - (rc.top + rc.bottom) / 2);

	SetWindowPos(window, NULL, rc.left, rc.top, rc.right - rc.left,
	             rc.bottom - rc.top, SWP_NOZORDER | SWP_NOACTIVATE);
}

static void HandleGetDpiScaledSize(HWND window, UINT dpi, SIZE *size)
{
	// Windows 10 sends WM_GETDPISCALEDSIZE to get the application's preferred
	// size before it changes to the specified DPI. Oni can't change resolution
	// while it's running, so the preferred size is always the current physical
	// size with added borders.
	RECT rc;
	GetClientRect(window, &rc);
	DDrAdjustWindowRect(&rc, dpi);
	size->cx = rc.right - rc.left;
	size->cy = rc.bottom - rc.top;
}

static void HandlePaint(HWND window)
{
	PAINTSTRUCT ps;
	HDC dc = BeginPaint(window, &ps);

	// Oni does a useless PatBlt here
	// PatBlt(dc, 0, 0, 4096, 4096, BLACKNESS);

	EndPaint(window, &ps);
}

static void HandleMouseWheel(POINTS pts, WORD key_state, short wheel_delta)
{
	static int delta_acc = 0;
	static int64_t delta_updated = 0;

	int64_t now = UUrMachineTime_High();
	int64_t last_updated = now - delta_updated;
	delta_updated = now;

	// Reset the accumulator if too much time has passed since the last scroll
	// event. The user is assumed to have finished scrolling.
	if (last_updated / UUrMachineTime_High_Frequency() > 0.3)
		delta_acc = 0;

	// Get the number of whole lines scrolled, if any. To support touchpad
	// scrolling and mice with continuous scroll wheels, the wheel delta is
	// accumulated and an input event is generated when at least one line is
	// scrolled. The accumulated wheel delta is stored pre-multiplied with
	// WheelScrollLines so it doesn't lose unnecessary precision.
	delta_acc += wheel_delta * DDgPlat_WheelScrollLines;
	int lines = delta_acc / WHEEL_DELTA;
	delta_acc = delta_acc % WHEEL_DELTA;
	if (!lines)
		return;

	// WM_MOUSEWHEEL events are in screen space (for some reason?)
	POINT pt = { pts.x, pts.y };
	ScreenToClient(ONgPlatformData.Window, &pt);
	IMtPoint ipt = { pt.x, pt.y };

	// Flip the sign of lines because (imo) it makes more sense for negative
	// values to scroll up and positive values to scroll down.
	LIrInputEvent_Add(DDcEventType_MouseWheel, &ipt, -lines, key_state);
}

static bool HandleSetCursor(WORD hittest)
{
	// If a mouse is inside our client area, we hide cursor (always), otherwise
	// we ask DefWindowProc to set an appropriate arrow for us.
	if (hittest == HTCLIENT) {
		SetCursor(NULL);
		return true;
	}
	return false;
}

static LRESULT CALLBACK DD_ONrPlatform_WindowProc(HWND window, UINT msg, WPARAM wparam, LPARAM lparam)
{
	LRESULT res;
	if (DDrInput_WindowProc(window, msg, wparam, lparam, &res))
		return res;

	switch (msg) {
	case WM_CLOSE:
		HandleClose();
		return 0;
	case WM_DPICHANGED:
		HandleDpiChanged(window, HIWORD(wparam), (RECT *)lparam);
		return 0;
	case WM_GETDPISCALEDSIZE:
		HandleGetDpiScaledSize(window, wparam, (SIZE *)lparam);
		return TRUE;
	case WM_PAINT:
		HandlePaint(window);
		return 0;
	case WM_MOUSEWHEEL:
		HandleMouseWheel(MAKEPOINTS(lparam), GET_KEYSTATE_WPARAM(wparam),
		                 GET_WHEEL_DELTA_WPARAM(wparam));
		break;
	case WM_SYSCOMMAND:
		// Prevent screen saver from starting when Oni has focus.
		if (wparam == SC_SCREENSAVE)
			return 0;
		break;
	case WM_SETCURSOR:
		if (HandleSetCursor(LOWORD(lparam)))
			return TRUE;
		break;
	}

	return ONrPlatform_WindowProc(window, msg, wparam, lparam);
}

UUtError ONICALL DD_ONrPlatform_Initialize(ONtPlatformData *PlatformData)
{
 	WNDCLASSEX WndClass;
 	RECT Rect;
 	const int Width = 640, Height = 480;
	DWORD window_style, window_style_ex;
 
	PlatformData->Instance = ONgInstance;
	PlatformData->Window   = NULL;
	
	if (FindWindow("ONI ", "ONI "))
	{
		AUrMessageBox(1, "Daodan: There is already an instance of the game running.");
		exit(0);
	}

	if (DDgDpiAwareness == DPI_AWARENESS_CONTEXT_PER_MONITOR_AWARE_V2) {
		HMODULE user32 = GetModuleHandleW(L"user32.dll");
		pAdjustWindowRectExForDpi = (void *)GetProcAddress(user32, "AdjustWindowRectExForDpi");
		pGetDpiForWindow = (void *)GetProcAddress(user32, "GetDpiForWindow");
	}

	UINT scroll_lines = 3;
	SystemParametersInfoA(SPI_GETWHEELSCROLLLINES, 0, &scroll_lines, 0);
	DDgPlat_WheelScrollLines = scroll_lines;

	WndClass.cbSize        = sizeof(WndClass);
	WndClass.style         = CS_VREDRAW | CS_HREDRAW | CS_OWNDC;
	WndClass.cbClsExtra    = 0;
	WndClass.cbWndExtra    = 0;
	WndClass.hInstance     = PlatformData->Instance;
	WndClass.hCursor       = LoadCursor(NULL, IDC_ARROW);
	WndClass.hIcon         = LoadIcon(ONgInstance, MAKEINTRESOURCE(103));
	WndClass.hIconSm       = LoadIcon(ONgInstance, MAKEINTRESOURCE(103));
	WndClass.hbrBackground = GetStockObject(BLACK_BRUSH);
	WndClass.lpszMenuName  = NULL;
 	WndClass.lpszClassName = "ONI ";
	WndClass.lpfnWndProc   = DD_ONrPlatform_WindowProc;
 
 	RegisterClassEx(&WndClass);
 	
	if (M3gResolutionSwitch)
	{
		// Do not allow border and topmost flag for a fullscreen window.
		window_style    = WS_POPUP;
		window_style_ex = 0;
	}
	else
	{
		ConfigOption_t* co = DDrConfig_GetOptOfType("windows.border", C_BOOL);
		int opt_border = 0;
		if (co && co->value.intBoolVal)
			opt_border = 1;
		co = DDrConfig_GetOptOfType("windows.topmost", C_BOOL);
		int opt_topmost = 0;
		if (co && co->value.intBoolVal)
			opt_topmost = 1;

		window_style    = (opt_border) ? WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU | WS_DLGFRAME | WS_MINIMIZEBOX : WS_POPUP;
		window_style_ex = (opt_topmost) ? WS_EX_TOPMOST : 0;
	}
	
 	Rect.left = (GetSystemMetrics(SM_CXSCREEN) / 2) - (Width / 2);
 	Rect.top = (GetSystemMetrics(SM_CYSCREEN) / 2) - (Height / 2);
 	Rect.right = Rect.left + Width;
 	Rect.bottom = Rect.top + Height;
	AdjustWindowRectEx(&Rect, window_style, FALSE, window_style_ex);
	
	PlatformData->Window = CreateWindowEx(window_style_ex, WndClass.lpszClassName, "ONI ", window_style,
		Rect.left, Rect.top, Rect.right - Rect.left, Rect.bottom - Rect.top,
		NULL, NULL, PlatformData->Instance, NULL);
	
	ShowWindow(PlatformData->Window, SW_SHOWNORMAL);
	UpdateWindow(PlatformData->Window);

	return UUcError_None;
}
