#include "Input.h"
#include "WindowManager.h"

UUtBool DDrWindow_GetRectPure(WMtWindow *window, IMtRect *rect)
{
	// Get the window position in its parent's client coordinates. Unclear
	// whether this was just never called in Windows Oni, or if it was written
	// just for Intel Mac Oni 1.2.1, but this doesn't exist in the Windows Oni
	// binary. Mac 1.2.1 uses it to get the position of the main menu resume
	// button, so it can be moved relative to its old position.

	if (!WMiWA_Valid(window))
		return false;

	*rect = (IMtRect) {
		.left = window->left,
		.top = window->top,
		.right = window->left + window->width,
		.bottom = window->top + window->height,
	};
	return true;
}

static void DDiTimer_Update(void)
{
	// Oni's WMiTimer_Update. This checks the list of timers for expiry and
	// fires off their window messages (think like SetTimer in Win32.)
	// Unfortunately, this was inlined in WMrUpdate, so now it has to be here.

	uint32_t now = UUrMachineTime_Sixtieths();
	WMtTimer *timers = UUrMemory_Array_GetMemory(WMgTimers);
	size_t timer_count = UUrMemory_Array_GetUsedElems(WMgTimers);

	for (int i = 0; i < timer_count; i++) {
		WMtTimer *timer = &timers[i];

		// Skip timers that haven't yet elapsed
		if (timer->next_expiry >= now)
			continue;

		// Reset the timer. Oni's timers are periodic, so this will fire again
		// when the period next expires.
		timer->next_expiry = now + timer->period;

		// Fire the timer message. These timers send messages to windows, like
		// SetTimer and WM_TIMER in Win32
		WMrMessage_Send(timer->window, WMcWM_Timer, timer->event_id, 0);

		// Check if the timer array has been modified. If it has, the array data
		// may be invalid, so we should reload it and start again. We go one
		// better than Oni here and check if either the timer array memory or
		// the timer count has been modified, rather than just the timer count.
		WMtTimer *new_timers = UUrMemory_Array_GetMemory(WMgTimers);
		uint32_t new_timer_count = UUrMemory_Array_GetUsedElems(WMgTimers);
		if (new_timers != timers || timer_count != new_timer_count) {
			timers = new_timers;
			timer_count = new_timer_count;
			i = 0;
		}
	}
}

static WMtWindow *DDrGetWindowUnderPoint(IMtPoint *point)
{
	// Oni's WMrGetWindowUnderPoint. Inlined in Windows Oni.
	return WMiWindow_GetWindowUnderPoint(WMgDesktop, point,
	                                     WMcWindowState_Visible);
}

static void DDiEvents_HandleMouseEvent(LItInputEvent *event)
{
	WMgCursorPosition = event->mouse_pos;

	WMtWindow *target = NULL;
	if (WMgMouseFocus) {
		// WMgMouseFocus is the window with mouse capture, which overrides the
		// window under the cursor
		target = WMgMouseFocus;
	} else {
		// Otherwise mouse messages, including our new scroll messages, go to
		// the window other the cursor
		target = DDrGetWindowUnderPoint(&event->mouse_pos);

		// Windows that are disabled, or have a second mystery state, don't get
		// mouse events
		if (target && (target->state & (WMcWindowState_Disabled |
		                                WMcWindowState_Unknown)))
			target = NULL;
	}

	// The first message parameter is the x/y coords of the mouse event
	uint32_t param1 = (uint32_t)event->mouse_pos.x << 16 |
	                  (uint16_t)event->mouse_pos.y;
	// The second parameter of a normal message is the key modifiers, but this
	// can be overridden by special messages
	uint32_t param2 = event->modifiers;

	// Did we hit a non-client area of the window?
	uint32_t hittest = WMcHitTest_Client;
	if (target)
		hittest = WMrMessage_Send(target, WMcWM_NC_HitTest, param1, 0);

	int msg;
	if (hittest == WMcHitTest_Client) {
		int client_msgs[] = {
			[LIcEventType_MouseMove] = WMcWM_MouseMove,
			[LIcEventType_LButtonDown] = WMcWM_LButtonDown,
			[LIcEventType_LButtonUp] = WMcWM_LButtonUp,
			[LIcEventType_MButtonDown] = WMcWM_MButtonDown,
			[LIcEventType_MButtonUp] = WMcWM_MButtonUp,
			[LIcEventType_RButtonDown] = WMcWM_RButtonDown,
			[LIcEventType_RButtonUp] = WMcWM_RButtonUp,
			[DDcEventType_MouseWheel] = DDcWM_MouseWheel,
		};
		msg = client_msgs[event->type];
	} else {
		int nonclient_msgs[] = {
			[LIcEventType_MouseMove] = WMcWM_NC_MouseMove,
			[LIcEventType_LButtonDown] = WMcWM_NC_LButtonDown,
			[LIcEventType_LButtonUp] = WMcWM_NC_LButtonUp,
			[LIcEventType_MButtonDown] = WMcWM_NC_MButtonDown,
			[LIcEventType_MButtonUp] = WMcWM_NC_MButtonUp,
			[LIcEventType_RButtonDown] = WMcWM_NC_RButtonDown,
			[LIcEventType_RButtonUp] = WMcWM_NC_RButtonUp,
			[DDcEventType_MouseWheel] = DDcWM_MouseWheel, // No NC wheel event
		};
		msg = nonclient_msgs[event->type];

		// For non-client events, the second message parameter is the hittest
		param2 = hittest;

		// No WMcWM_NC_MouseMove when the mouse is captured
		if (WMgMouseFocus && msg == WMcWM_NC_MouseMove)
			msg = WMcWM_MouseMove;
	}

	// We smuggle the wheel lines in the high 16-bits of param2
	if (msg == DDcWM_MouseWheel)
		param2 |= (uint32_t)event->ch << 16;

	// Changes the window message ID to the double-click equivalent
	int dblclk_msg = msg;
	if (WMiEvent_IsDoubleClick(event, target, &dblclk_msg))
		WMrMessage_Post(target, dblclk_msg, param1, param2);

	// Post the actual message to the window's message queue
	WMrMessage_Post(target, msg, param1, param2);
}

static void DDiEvents_Process(void)
{
	// Oni's WMiEvents_Process. This was also inlined in WMrUpdate.

	if (LIrMode_Get())
		return;

	for (int i = 0; i < 10; i++) {
		// Note that Oni kept looping even when LIrInputEvent_Get returned
		// false. We break instead, because there's no more input events to
		// process.
		LItInputEvent event;
		if (!LIrInputEvent_Get(&event))
			break;

		// Clip mouse position to the desktop size
		if (event.mouse_pos.x < 0)
			event.mouse_pos.x = 0;
		if (event.mouse_pos.y < 0)
			event.mouse_pos.y = 0;
		if (WMgDesktop) {
			short desktop_width;
			short desktop_height;
			WMrWindow_GetSize(WMgDesktop, &desktop_width, &desktop_height);
			if (event.mouse_pos.x > desktop_width)
				event.mouse_pos.x = desktop_width;
			if (event.mouse_pos.y > desktop_height)
				event.mouse_pos.y = desktop_height;
		}

		switch (event.type) {
		case LIcEventType_MouseMove:
		case LIcEventType_LButtonDown:
		case LIcEventType_LButtonUp:
		case LIcEventType_MButtonDown:
		case LIcEventType_MButtonUp:
		case LIcEventType_RButtonDown:
		case LIcEventType_RButtonUp:
		case DDcEventType_MouseWheel:
			DDiEvents_HandleMouseEvent(&event);
			break;
		case LIcEventType_KeyPress:
		case LIcEventType_UnknownKeyboard:
			WMiEvents_HandleKeyEvent(&event);
			break;
		}
	}
}

void ONICALL DD_WMrUpdate(void)
{
	if (!WMgActive)
		return;
	DDiTimer_Update();
	DDiEvents_Process();
	SS2rUpdate();
}

static void DDiListBox_HandleMouseWheel(WMtWindow *window, short lines)
{
	// Scroll up
	for (int i = lines; i < 0; i++)
		WMiListBox_HandleVerticalScroll(window, 0, 1);

	// Scroll down
	for (int i = lines; i > 0; i--)
		WMiListBox_HandleVerticalScroll(window, 0x10000, 1);
}

uint32_t ONICALL DDiListBox_Callback(WMtWindow *window, int msg,
                                     uint32_t param1, uint32_t param2)
{
	switch (msg) {
	case DDcWM_MouseWheel:
		DDiListBox_HandleMouseWheel(window, param2 >> 16);
		return 0;
	}

	return WMiListBox_Callback(window, msg, param1, param2);
}
